/*
 * usage:  ppmview <imagefile>
 *
 * displays on the screen a PGM or PPM image (raw mode only).
 */

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>

#include "gl\gl.h"
#include "gl\glu.h"
#include "gl\glaux.h"

#define USE_ISPRINTF		LPSTR ___isprintfbuf=(char*)_alloca(1024)
#define __                  (sprintf(___isprintfbuf,
#define _                   ),___isprintfbuf)

long WINAPI WndProc( HWND hWnd, WORD message, WORD wParam, LONG lParam );
BOOL LoadPPM( LPSTR filename );

LPSTR message = NULL;
int pixdatatype;
int width;
int height;
unsigned char* data;

int WINAPI WinMain( HANDLE hInstance, HANDLE hPrevInstance, LPSTR lpszCmdParam,
			int nCmdShow )
{
	USE_ISPRINTF;
	static char		szAppName[] = "PPMView";
	HWND			hWnd;
	MSG				msg;
	WNDCLASS		wndclass;

	if ( !hPrevInstance )
	{
		wndclass.style = CS_HREDRAW | CS_VREDRAW;
		wndclass.lpfnWndProc = (WNDPROC)WndProc;
		wndclass.cbClsExtra = 0;
		wndclass.cbWndExtra = 0;
		wndclass.hInstance = hInstance;
		wndclass.hIcon = LoadIcon( NULL, IDI_APPLICATION );
		wndclass.hCursor = LoadCursor( NULL, IDC_ARROW );
		wndclass.hbrBackground = NULL;
		wndclass.lpszMenuName = NULL;
		wndclass.lpszClassName = szAppName;

		RegisterClass( &wndclass );
	}

	LoadPPM( lpszCmdParam );
	
	hWnd = CreateWindowEx( 0,
		szAppName, 
		data ? __ "%s (%d x %d)", lpszCmdParam, width, height _ :
	           message,
		WS_OVERLAPPED | WS_SYSMENU | WS_CLIPSIBLINGS | WS_CLIPCHILDREN, 
		CW_USEDEFAULT, 
		CW_USEDEFAULT, 
		data ? width + 2 * GetSystemMetrics( SM_CXFIXEDFRAME ) : 400,
		data ? height + 2 * GetSystemMetrics( SM_CYFIXEDFRAME ) + 
			GetSystemMetrics( SM_CYCAPTION ) : 100, 
		NULL, 
		NULL, 
		hInstance, 
		NULL );

	ShowWindow( hWnd, nCmdShow );
	UpdateWindow( hWnd );

	while( GetMessage( &msg, NULL, 0, 0 ) )
	{
		TranslateMessage( &msg );
		DispatchMessage( &msg );
	}

	return msg.wParam;
}

HGLRC InitializeOpenGL( HWND hWnd )
{
	HDC g_hdc = GetDC( hWnd );

	PIXELFORMATDESCRIPTOR pfd = {
		sizeof(PIXELFORMATDESCRIPTOR),
		1,
		PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
		PFD_TYPE_RGBA,
		24,
		0, 0, 0, 0, 0, 0,
		0,
		0,
		0,
		0, 0, 0, 0,
		16,
		0,
		0,
		PFD_MAIN_PLANE,
		0,
		0, 0, 0
	};

	int pixelformat;
	if ( ( pixelformat = ChoosePixelFormat( g_hdc, &pfd ) ) == 0 )
		return NULL;

	if ( SetPixelFormat( g_hdc, pixelformat, &pfd ) == 0 )
		return NULL;

	HGLRC hrc;

	if ( ( hrc = wglCreateContext( g_hdc ) ) == 0 )
		return NULL;

	ReleaseDC( hWnd, g_hdc );

	return hrc;
}

long WINAPI WndProc( HWND hWnd, WORD message, WORD wParam, LONG lParam )
{
	USE_ISPRINTF;

	int crw, crh;
	HDC hdc;
	PAINTSTRUCT ps;
	RECT rect;
	static HGLRC hrc = NULL;

	switch( message )
	{

	case WM_CREATE:
		hrc = InitializeOpenGL( hWnd );
		return 0;

	case WM_NCPAINT:
		OutputDebugString( "WM_NCPAINT\n" );
		break;

	case WM_ERASEBKGND:
		OutputDebugString( "WM_ERASEBKGND\n" );
		return 1;

	case WM_PAINT:
		OutputDebugString( "WM_PAINT\n" );

		if ( GetUpdateRect( hWnd, NULL, false ) == 0 )
			return 0;

		hdc = BeginPaint( hWnd, &ps );

		crw = GetClientRect( hWnd, &rect );
		crh = GetLastError();

		crw = rect.right;
		crh = rect.bottom;
		OutputDebugString( __ "client rect is %d x %d\n", crw, crh _ );

		wglMakeCurrent( hdc, hrc );

		glViewport( 0, 0, crw, crh );

		glMatrixMode( GL_PROJECTION );
		glLoadIdentity();
		gluOrtho2D( 0, crw, 0, crh );

		glMatrixMode( GL_MODELVIEW );
		glLoadIdentity();
		glTranslatef( 0.0f, (float)(crh - height), 0.0 );
		glTranslatef( 0.375, 0.375, 0.0 );

		glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
		glClear( GL_COLOR_BUFFER_BIT );

		if ( data )
		{
			glRasterPos2i( 0, 0 );
			glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
			glPixelStorei( GL_UNPACK_ROW_LENGTH, width );
			glDrawPixels( width, height, pixdatatype, GL_UNSIGNED_BYTE, data );
		}

		glFinish();
		SwapBuffers( hdc );

		wglMakeCurrent( NULL, NULL );

		EndPaint( hWnd, &ps );
		return 0;

	case WM_DESTROY:
		PostQuitMessage( 0 );
		return 0;
	}

	return DefWindowProc( hWnd, message, wParam, lParam );
}

BOOL LoadPPM( LPSTR filename )
{
	USE_ISPRINTF;
	FILE* f;
	int i, j;
	char buffer[80];
	int phase = 0;
	int type;
	int maxval;

	f = fopen( filename, "rb" );
	if ( f == NULL )
	{
		message = strdup( __ "couldn't open \"%s\"", filename _ );
		data = NULL;
		return false;
	}
	
	while(phase < 4)
	{
		fgets( buffer, 80, f );
		if ( buffer[0] == '#' ) continue;

		switch( phase )
		{
		case 0:	j = sscanf( buffer, "P%d %d %d %d\n", &type, &width, &height, &maxval ); break;
		case 1:	j = sscanf( buffer, "%d %d %d\n", &width, &height, &maxval ); break;
		case 2:	j = sscanf( buffer, "%d %d\n", &height, &maxval ); break;
		case 3:	j = sscanf( buffer, "%d\n", &maxval ); break;
		case 4: j = 0; break;
		}

		phase += j;
	}

	OutputDebugString( __ "type is %d; picture is %d x %d; maxval is %d\n", type, width, height, maxval _ );

	switch ( type )
	{
	case 5:
		data = (unsigned char*)malloc( width * height );
		for ( i = 0; i < height; ++i )
			fread( data + (height-1-i)*width, 1, width, f );
		pixdatatype = GL_LUMINANCE;
		break;

	case 6:
		data = (unsigned char*)malloc( width * height * 3 );
		for ( i = 0; i < height; ++i )
			fread( data + (height-1-i)*width*3, 1, width * 3, f );
		pixdatatype = GL_RGB;
		break;

	default:
		message = strdup( __ "unknown PPM type %d", type _ );
		data = NULL;
	}

	fclose( f );

	return true;
}

